'' =================================================================================================
''
''   File....... jm_ws2801.spin
''   Purpose.... High-performance driver for the WS2801
''   Author..... Jon "JonnyMac" McPhalen 
''               Copyright (c) 2012 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 28 SEP 2012
''
'' =================================================================================================

{

  Recommended frame timing (@ 80MHz) is: (10us x channels) + 1ms

}


con

  RX1 = 31                                                      ' programming / terminal
  TX1 = 30
  
  SDA = 29                                                      ' eeprom / i2c
  SCL = 28

  
con

  MAX_CHANNELS = 512 / 3 { 170 }                                ' one dmx universe


con

  RED    = $FF_00_00
  GREEN  = $00_FF_00
  BLUE   = $00_00_FF

  YELLOW = $FF_FF_00
  CYAN   = $00_FF_FF
  VIOLET = $FF_00_FF

  WHITE  = $FF_FF_FF
  BLACK  = $00_00_00
  

var

  long  cog                                                     ' id of cog running driver

  long  dpin                                                    ' -> SDI pin
  long  cpin                                                    ' -> CKI pin
  long  chcount                                                 ' channel count in frame
  long  frametix                                                ' frame timing (in cnt ticks)
  long  rgbfix                                                  ' red/blue exchange flag
  long  grbfix                                                  ' red/green exchange flag
  long  lpdfix                                                  ' LPD convert flag
  
  long  color[MAX_CHANNELS]                                     ' color array


pub start(sdi, cki, ch, ms, rbswap, rgswap, LPDconvert)

'' Start multi-channel WS2801 driver
'' -- sdi is serial data pin
'' -- cki is clock pin
'' -- ch is the number of channels (RGB groups) in string
'' -- ms is the frame refresh timing in milliseconds
'' -- rbswap is flag to swap red and blue channels for output  

  stop

  dpin     := sdi                                               ' set pins
  cpin     := cki
  chcount  := 1 #> ch <# MAX_CHANNELS                           ' set channels (w/limits)
  frametix := clkfreq / 1_000 * ms                              ' ticks in frame
  rgbfix   := (rbswap <> 0)                                     ' exchange r and b bytes in color
  grbfix   := (rgswap <> 0)                                     ' exchange r ang g bytes in colod
  lpdfix   := (LPDconvert <> 0)                                 ' convert from a LPD strip
  
  cog := cognew(@ws2801, @dpin) + 1                             ' start the cog

  return cog


pub stop

'' Stops cog (if running)

  if (cog)
    off
    waitcnt(clkfreq/100 +cnt)
    cogstop(cog - 1)
    cog := 0

  longfill(@color, $00_00_00, MAX_CHANNELS)                     ' clear all to off
  

pub set(ch, rgb) | work

'' Sets channel to RGB value
'' -- re-orders bytes to accomodate RGB value (e.g., $FF_FF_00 = yellow)

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    color[ch] := rgb & $FF_FF_FF                                '  update color array


pub set_red(ch, level)

'' Sets red led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 2] := 0 #> level <# 255           '  set it
    

pub set_green(ch, level)

'' Sets green led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 1] := 0 #> level <# 255           '  set it


pub set_blue(ch, level)

'' Sets blue led level of selected channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    byte[@color + (ch << 2) + 0] := 0 #> level <# 255           '  set it

    
pub set_all(rgb) | work

'' Sets all leds in string to RGB
  
  longfill(@color, rgb & $FF_FF_FF, chcount)


pub set_range(first, last, rgb) | work

'' Sets range of LEDs in string to RGB

  first := 0 #> first <# chcount-1                              ' keep first channel legal
  last := 0 #> last <# chcount-1                                ' keep last channel legal
  rgb &= $FF_FF_FF                                              ' keep color legal

  if (first > last)                                             ' need to swap first and last
    work := last
    last := first
    first := work

  longfill(@color[first], rgb, last-first+1)                    ' update array     

  
pub off

'' Turns off all LEDs

  longfill(@color, $00_00_00, chcount)


pub make_rgb(r, g, b)

'' Combine R, G, and B color values into 24-bits
'' GRB

  'return ((($00 #> g <# $FE)+1) << 15) | ((($00 #> r <# $FE)+1) << 7) | (($00 #> b <# $FE)+1)>>1
  return (($00 #> r <# $FF) << 16) | (($00 #> g <# $FF) << 8) | ($00 #> b <# $FF)
               

pub read(ch)

'' Returns color of channel

  if ((ch => 0) and (ch < MAX_CHANNELS))                        ' valid?
    return color[ch]
  else
    return 0

  
pub address

'' Returns address of color array

  return @color
  

pub running

'' Returns > 0 if running

  return cog
 

dat

                        org     0

ws2801                  mov     t1, par                         ' copy hub address of parameters
                        rdlong  sdipin, t1                      ' read sdi pin #
                        add     t1, #4
                        rdlong  ckipin, t1                      ' read cki pin #
                        add     t1, #4
                        rdlong  nleds, t1                       ' read # of channels
                        add     t1, #4
                        rdlong  frame, t1                       ' read frame timing
                        add     t1, #4
                        rdlong  xrb, t1                         ' read red/blue exchange flag
                        add     t1, #4
                        rdlong  xrg, t1                         ' read red/blue exchange flag
                        add     t1, #4
                        rdlong  xlpd, t1                         ' read red/blue exchange flag
                        add     t1, #4                                                
                        mov     hub, t1                         ' save hub address of colors

                        mov     sdimask, #1                     ' convert sdi pin to mask
                        shl     sdimask, sdipin
                        or      dira, sdimask                   ' make output

                        mov     ckimask, #1                     ' convert cki pin to mask
                        shl     ckimask, ckipin
                        or      dira, ckimask                   ' make output

                        mov     frametimer, frame               ' setup frame timer
                        add     frametimer, cnt                 ' sync with system counter
                        
rgbmain                 mov     pntr, hub                       ' point to color[0]
                        mov     count, nleds                    ' set channels count

getrgb                  rdbyte  t2, pntr                        ' read rgb color value
                        add     pntr, #1                        ' point to next rgb value

                        mov     t1, t2
                        shl     t1, #8
                        
                        rdbyte  t2, pntr                        ' read rgb color value
                        add     pntr, #1                        ' point to next rgb value
                        
                        or      t1, t2
                        shl     t1, #8
                        
                        rdbyte  t2, pntr                        ' read rgb color value
                        add     pntr, #1                        ' point to next rgb value
                        
                        or      t1, t2

                        ' swap red and blue channels
                        ' -- swaps only if xrb <> 0

swaprb                  tjz     xrb, #swaprg                    ' skip if swap flag is 0    
' RGB -> BGR
                        mov     t2, t1                          ' copy original             
                        and     t2, HX_00_FF_00                 ' isolate green
                        
                        mov     t3, t1                          ' copy original
                        and     t3, HX_00_00_FF                 ' isolate blue
                        shl     t3, #16                         ' shift to correct position
                        or      t2, t3                          ' mix blue and green
                        
                        and     t1, HX_FF_00_00                 ' isolate red
                        shr     t1, #16                         ' shift to correct position
                        or      t1, t2                          ' mix blue, green, and red

swaprg                  tjz     xrg, #convLPD                   'skip if swaprg flag is 0
' RGB -> GRB
                        mov     t2, t1                          ' copy original
                        and     t2, HX_00_00_FF                 ' isolate blue

                        mov     t3, t1                          ' copy original
                        and     t3, HX_00_FF_00                 ' isolate green
                        shl     t3, #8                          ' shift to correct position
                        or      t2, t3                          ' mix blue and green

                        and     t1, HX_FF_00_00                 ' isolate red           
                        shr     t1, #8                          ' shift to correct position
                        or      t1, t2                          ' mix blue, green, and red

convLPD                 tjz     xlpd, #shiftout                 'skip if convertLPD flag is 0
' convert 8 bit color to 7 bit color
                        mov     t2, t1                          ' copy original
                        and     t2, HX_FF_00_00                 ' isolate green
                        max     t2, green_limit                 ' Limit max to 254 (shifted left to green position)
                        add     t2, green_shift                 ' add 1 (shifted left to green position)
                        
                        mov     t3, t1                          ' copy original
                        and     t3, HX_00_FF_00                 ' isolate red
                        max     t3, red_limit                   ' Limit max to 254 (shifted left to red position)
                        add     t3, red_shift                   ' add 1 (shifted left to red position)
                        or      t2, t3                          ' mix blue and green
                        
                        and     t1, HX_00_00_FF                 ' isolate blue
                        max     t1, blue_limit                  ' Limit max to 254 (shifted left to blue position)
                        add     t1, blue_shift                         ' add 1
                        or      t1, t2                          ' mix blue and green
                        
                        shr     t1, #1                          ' shift right 1
                        or      t1, LPD_or_mask                 ' set bit 7 for all colors
                        
                        ' shift bits to WS2801 chain        

shiftout                shl     t1, #8                          ' shift rgb msb to bit31
                        mov     bits, #24                       ' 24 bits = 3 bytes (r+g+b)
:loop                   rcl     t1, #1                  wc      ' move msb to C
                        muxc    outa, sdimask                   ' copy C to sdi 
                        nop                                     ' let sdi stabilize
                        or      outa, ckimask                   ' clock high
                        nop
                        andn    outa, ckimask                   ' clock low
                        djnz    bits, #:loop                    ' update bit count, do next
                        
                        djnz    count, #getrgb                  ' update count, get next

                        tjz     xlpd, #framewait                 'skip if convertLPD flag is 0    
                        mov     bits, #8                        ' shift out 1 "latch" byte for LPD strip
:floop                  andn    outa, sdimask
                        nop
                        or      outa, ckimask                   ' clock high
                        nop
                        andn    outa, ckimask                   ' clock low                          
                        djnz    bits, #:floop                    ' update bit count, do next
                         
framewait               waitcnt frametimer, frame               ' let frame finish, restart
                        jmp     #rgbmain

' --------------------------------------------------------------------------------------------------

HX_FF_00_00             long    $FF_00_00                       ' red byte mask
HX_00_FF_00             long    $00_FF_00                       ' green byte mask
HX_00_00_FF             long    $00_00_FF                       ' blue byte mask

green_limit             long    %11111110_00000000_00000000
red_limit               long    %11111110_00000000
blue_limit              long    %11111110
green_shift             long    %1_00000000_00000000
red_shift               long    %1_00000000
blue_shift              long    %1
LPD_or_mask             long    %10000000_10000000_10000000

sdipin                  res     1                               ' SDI (serial data) pin
ckipin                  res     1                               ' CKI (clock) pin
nleds                   res     1                               ' number of RGB leds in chain
frame                   res     1                               ' frame timing
xrb                     res     1                               ' red/blue exchange flag
xrg                     res     1                               ' red/green exchange flag
xlpd                    res     1                               ' LPD convert flag
hub                     res     1                               ' hub address of color[] array

sdimask                 res     1                               ' pin mask for SDI
ckimask                 res     1                               ' pin mask for CKI
pntr                    res     1                               ' data pointer
count                   res     1                               ' count of leds to process
bits                    res     1                               ' bits in shiftout

frametimer              res     1                               ' for frame timing

t1                      res     1                               ' work vars
t2                      res     1
t3                      res     1

                        fit     496                                    

                        
dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}  